// Updates: 2006.01.13


/*
 * The "write" and "append" methods do not create the file if it does not exist.
 * It would be easy to do it, but there will be other problems:
 * - Solution 1: creating each time the file, even if it already exists. The 
 * problem occurs if the Midlet runs in an untrusted security domain, because 
 * the user will be prompted once more.
 * - Solution 2: checking if the file exists with the "exists" method. The 
 * problem occurs if the Midlet runs in an untrusted security domain, because 
 * the "exists" method requires "read" access, and therefore the user will be
 * prompted once more.
 */


import java.io.*;
import javax.microedition.io.*;
import javax.microedition.io.file.*;


/**
 * The FileIO class is used to handle files.<br>
 * Three methods, namely <code>read</code>, <code>write</code>, and
 * <code>append</code>, work in an asynchronous manner. Such methods return 
 * immediately and this class calls the corresponding callback method once the 
 * operation is achieved. This class uses a single Thread to fulfil all the
 * asynchronous operations. It is therefore not possible to launch a new request
 * before the previous one is finished.
 * 
 * @author Michel Deriaz
 */
public class FileIO implements Runnable {
  
  /** The operation finished successfully. */
  public static final int STATE_OK = 0;

  /**
   * A problem occured during the operation. One reason could be that the file
   * does not exist.
   */
  public static final int STATE_PROBLEM = 1;
  
  /** The class is not ready. */
  public static final int STATE_NOT_READY = 2;
  
  /** The given filename is <code>null</code>. */
  public static final int STATE_FILENAME_NULL = 3;
  
  /** The given content is <code>null</code>. */
  public static final int STATE_CONTENT_NULL = 4;

  
  static final int OPERATION_NOP = 0;    
  static final int OPERATION_READ = 1;
  static final int OPERATION_WRITE = 2;
  static final int OPERATION_APPEND = 3;
    
  int operation = OPERATION_NOP;
  FileConnection fc;
  InputStream is;
  OutputStream os;
  String filename;
  byte[] content;
  FileIOListener listener;
  
  
  /**
   * Constructs a new FileIO object.
   * @param listener the listener that contains the callback methods
   */  
  public FileIO(FileIOListener listener) {
    this.listener = listener;
  }


  /**
   * Creates a new directory. Directories in the specified path are not
   * recursively created and must be explicitly created before subdirectories
   * can be created. Returns <code>STATE_FILENAME_NULL</code> if the specified 
   * filename is <code>null</code>, <code>STATE_PROBLEM</code> if a problem 
   * occurs (like creating a directory that does already exist or specifying a
   * path to the new subdirectory that doesn't exist), and <code>STATE_OK</code>
   * if the new directory is created successfully.
   * @param filename the full filename
   * @return a state code that describes how the operation ended up
   */
  public int mkdir(String filename) {
    if (filename == null) return STATE_FILENAME_NULL;
    try {
      fc = (FileConnection)Connector.open("file://" + filename, Connector.WRITE);
      fc.mkdir();
    }
    catch (Exception e) {
      System.out.println("FileIO.mkdir: " + e);
      return STATE_PROBLEM;
    }
    return STATE_OK;
  }
  
 
  /**
   * Creates a new file. This method does not create any directories specified
   * in the file's path. Returns <code>STATE_FILENAME_NULL</code> if the
   * specified filename is <code>null</code>, <code>STATE_PROBLEM</code> if a
   * problem occurs (like creating a file that does already exist or specifying
   * a path that doesn't exist), and <code>STATE_OK</code> if the new file is
   * created successfully.
   * @param filename the full filename
   * @return a state code that describes how the operation ended up
   */ 
  public int create(String filename) {
    if (filename == null) return STATE_FILENAME_NULL;
    try {
      fc = (FileConnection)Connector.open("file://" + filename, Connector.WRITE);
      fc.create();
    }
    catch (Exception e) {
      System.out.println("FileIO.create: " + e);
      return STATE_PROBLEM;
    }
    return STATE_OK;
  }
  
  
  /**
   * Checks if the specified file or directory exists. Returns
   * <code>STATE_FILENAME_NULL</code> if the specified filename is
   * <code>null</code>, <code>STATE_PROBLEM</code> if a problem occurs or if the
   * specified file or directory does not exist, and <code>STATE_OK</code> if
   * the file or directory exists.
   * @param filename the full filename   
   * @return a state code that describes how the operation ended up
   */
  public int exists(String filename) {
    if (filename == null) return STATE_FILENAME_NULL;       
    try {
      fc = (FileConnection)Connector.open("file://" + filename, Connector.READ);
      if (!fc.exists()) return STATE_PROBLEM;
    }
    catch (Exception e) {
      System.out.println("FileIO.exists: " + e);
      return STATE_PROBLEM;
    }
    return STATE_OK;
  }
      
  
  /**
   * Reads the specified file. This method returns immediately. The
   * <code>fileRead</code> callback method is called when the operation ends up.
   * @param filename the full filename
   * @see FileIOListener#fileRead(int state, byte[] content);
   */  
  public void read(String filename) {
    if (operation != OPERATION_NOP) listener.fileRead(STATE_NOT_READY, null);
    this.filename = filename;
    if (filename == null) listener.fileRead(STATE_FILENAME_NULL, null);
    else start(OPERATION_READ);
  }


  /**
   * Writes in the specified file. All former content is erased. This method
   * returns immediately. The <code>fileWritten</code> callback method is called
   * when the operation ends up.
   * @param filename the full filename
   * @param content the bytes to write
   * @see FileIOListener#fileWritten(int state);
   */  
  public void write(String filename, byte[] content) {
    if (operation != OPERATION_NOP) listener.fileWritten(STATE_NOT_READY);
    this.filename = filename;
    this.content = content;
    if (filename == null) listener.fileWritten(STATE_FILENAME_NULL);
    else if (content == null) listener.fileWritten(STATE_CONTENT_NULL);
    else start(OPERATION_WRITE);
  }
    

  /**
   * Appends the specified bytes to the specified file. This method returns
   * immediately. The <code>fileAppended</code> callback method is called
   * when the operation ends up.
   * @param filename the full filename
   * @param content the bytes to append
   * @see FileIOListener#fileAppended(int state);
   */  
  public void append(String filename, byte[] content) {
    if (operation != OPERATION_NOP) listener.fileAppended(STATE_NOT_READY);
    this.filename = filename;
    this.content = content;
    if (filename == null) listener.fileAppended(STATE_FILENAME_NULL);
    else if (content == null) listener.fileAppended(STATE_CONTENT_NULL);
    else start(OPERATION_APPEND);
  }

      
  public void run() {
    if (operation == OPERATION_READ) {
      try {
        fc = (FileConnection)Connector.open("file://" + filename, Connector.READ);
        is = fc.openInputStream();
        content = new byte[(int)fc.fileSize()];
        is.read(content);
        listener.fileRead(STATE_OK, content);
      }
      catch (Exception e) {
        System.out.println("FileIO.read: " + e);
        listener.fileRead(STATE_PROBLEM, null);
      }
      finally {
        close();
      }
    }
    if (operation == OPERATION_WRITE) {
      try {
        fc = (FileConnection)Connector.open("file://" + filename, Connector.WRITE);
        os = fc.openOutputStream();
        os.write(content);
        fc.truncate(content.length); // flush is done automatically
        listener.fileWritten(STATE_OK);
      }
      catch (Exception e) {
        System.out.println("FileIO.write: " + e);
        listener.fileWritten(STATE_PROBLEM);
      }
      finally {
        close();
      }
    }
    if (operation == OPERATION_APPEND) {
      try {
        fc = (FileConnection)Connector.open("file://" + filename, Connector.WRITE);
        os = fc.openOutputStream(Integer.MAX_VALUE);
        os.write(content);
        os.flush();
        listener.fileAppended(STATE_OK);
      }
      catch (Exception e) {
        System.out.println("FileIO.append: " + e);
        listener.fileAppended(STATE_PROBLEM);
      }
      finally {
        close();
      }
    }    
  }
  
  
  private void start(int operation) {
    this.operation = operation;
    Thread runner = new Thread(this);
    runner.start();
  }
  

  private void close() {
    try {
      is.close();
    }
    catch (Exception e) {}
    try {
      os.close();
    }
    catch (Exception e) {}
    try {
      fc.close();
    }
    catch (Exception e) {}
    operation = OPERATION_NOP;
  }
}